import * as glMatrix from "gl-matrix";
function toCSSNumber(v) {
  if (Math.abs(v) < 1e-6) {
    return "0.000000";
  }
  const res = v.toFixed(6);
  if (res.indexOf("e") >= 0) {
    throw new Error("eeeee");
  }
  return res;
}
function pos(x, y) {
  return glMatrix.vec2.fromValues(x, y);
}
function posAdd(...poss) {
  return poss.reduce((p, c) => {
    p[0] += c[0];
    p[1] += c[1];
    return p;
  }, pos(0, 0));
}
function genMatrix(s1, s2, s3, t1, t2, t3) {
  const input = Float32Array.from([
    s1[0],
    s2[0],
    s3[0],
    s1[1],
    s2[1],
    s3[1],
    1,
    1,
    1
  ]);
  const inverted = glMatrix.mat3.create();
  glMatrix.mat3.invert(inverted, input);
  const target = Float32Array.from([
    t1[0],
    t2[0],
    t3[0],
    t1[1],
    t2[1],
    t3[1],
    1,
    1,
    1
  ]);
  const transformMatrix = glMatrix.mat3.create();
  glMatrix.mat3.mul(transformMatrix, inverted, target);
  return transformMatrix;
}
const zero = pos(0, 0);
function genReflectMatrix(p1, p2) {
  const normal = glMatrix.vec2.create();
  glMatrix.vec2.subtract(normal, p1, p2);
  glMatrix.vec2.rotate(normal, normal, zero, Math.PI / 2);
  glMatrix.vec2.normalize(normal, normal);
  const center = glMatrix.vec2.create();
  glMatrix.vec2.lerp(center, p1, p2, 0.5);
  const p3 = glMatrix.vec2.clone(center);
  glMatrix.vec2.add(p3, p3, normal);
  const p3r = glMatrix.vec2.clone(center);
  glMatrix.vec2.subtract(p3r, p3r, normal);
  return genMatrix(p1, p2, p3, p1, p2, p3r);
}
function toCSSMatrix(transformMatrix, unit) {
  const transposed2dMatrix = [
    transformMatrix[0],
    transformMatrix[3],
    transformMatrix[1],
    transformMatrix[4],
    transformMatrix[2],
    transformMatrix[5]
  ];
  const deform = `matrix(${[...transposed2dMatrix.slice(0, 4)].map(toCSSNumber).join(", ")},0,0)`;
  const move = `translate(calc(${unit} * ${toCSSNumber(
    transposed2dMatrix[4]
  )}), calc(${unit} * ${toCSSNumber(transposed2dMatrix[5])}))`;
  const transform = `${move} ${deform}`;
  return transform;
}
const line = (p1, p2) => {
  return [p1, p2];
};
function lineIntersect(x1, y1, x2, y2, x3, y3, x4, y4) {
  var ua, ub, denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1);
  if (denom == 0) {
    return null;
  }
  ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denom;
  ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / denom;
  return [
    x1 + ua * (x2 - x1),
    y1 + ua * (y2 - y1),
    ua >= 0 && ua <= 1,
    ub >= 0 && ub <= 1
  ];
}
function intersection(line1, line2) {
  const res = lineIntersect(
    line1[0][0],
    line1[0][1],
    line1[1][0],
    line1[1][1],
    line2[0][0],
    line2[0][1],
    line2[1][0],
    line2[1][1]
  );
  if (res) {
    return {
      pos: pos(res[0], res[1]),
      hitLine1: res[2],
      hitLine2: res[3]
    };
  } else {
    return null;
  }
}
function expandPolygon(points, edgeToExpand = [], amount = 1) {
  if (amount === 0) {
    return points;
  }
  const edges = [];
  for (let i = 0; i < points.length; i++) {
    edges.push(line(points[i], points[(i + 1) % points.length]));
  }
  const center = points.reduce((p, c) => {
    p[0] += c[0] / points.length;
    p[1] += c[1] / points.length;
    return p;
  }, glMatrix.vec2.create());
  const edge0 = edges[0];
  const edge0Vec = Float32Array.from([
    edge0[1][0] - edge0[0][0],
    edge0[1][1] - edge0[0][1]
  ]);
  const edgeToCenterVec = Float32Array.from([
    center[0] - edge0[0][0],
    center[1] - edge0[0][1]
  ]);
  const isClockWise = edge0Vec[0] * edgeToCenterVec[1] - edgeToCenterVec[0] * edge0Vec[1] < 0;
  const newEdges = [];
  for (let i = 0; i < edges.length; i++) {
    const old = edges[i];
    if (!edgeToExpand.includes(i)) {
      newEdges.push(old);
    } else {
      const normal = Float32Array.from([
        old[1][1] - old[0][1],
        old[0][0] - old[1][0]
      ]);
      glMatrix.vec2.normalize(normal, normal);
      glMatrix.vec2.scale(normal, normal, isClockWise ? -amount : amount);
      newEdges.push(
        line(
          glMatrix.vec2.add(glMatrix.vec2.create(), old[0], normal),
          glMatrix.vec2.add(glMatrix.vec2.create(), old[1], normal)
        )
      );
    }
  }
  const newPoints = [];
  for (let i = 0; i < newEdges.length; i++) {
    const prev = newEdges[(i - 1 + newEdges.length) % newEdges.length];
    const old = newEdges[i];
    if (glMatrix.vec2.equals(prev[1], old[0])) {
      newPoints.push(prev[1]);
    } else {
      const newPoint = intersection(prev, old);
      if (newPoint) {
        newPoints.push(newPoint.pos);
      } else {
        if (glMatrix.vec2.equals(prev[0], prev[1])) {
          const prevNew = line(prev[0], old[0]);
          const newPoint2 = intersection(prevNew, old);
          if (newPoint2) {
            newPoints.push(newPoint2.pos);
          } else {
            console.error("bad expand");
            return points;
          }
        } else if (glMatrix.vec2.equals(old[0], old[1])) {
          const oldNew = line(prev[1], old[1]);
          const newPoint2 = intersection(prev, oldNew);
          if (newPoint2) {
            newPoints.push(newPoint2.pos);
          } else {
            console.error("bad expand");
            return points;
          }
        } else {
          console.error("bad expand");
          return points;
        }
      }
    }
  }
  return newPoints;
}
function applyTransform(point, transformMatrix) {
  const transformInput = glMatrix.vec3.fromValues(...point, 1);
  const transformOutput = glMatrix.vec3.create();
  const flippedMatrix = glMatrix.mat3.create();
  glMatrix.mat3.transpose(flippedMatrix, transformMatrix);
  glMatrix.vec3.transformMat3(transformOutput, transformInput, flippedMatrix);
  return pos(transformOutput[0], transformOutput[1]);
}
function applyReverseTransform(point, transformMatrix) {
  const transformInput = glMatrix.vec3.fromValues(...point, 1);
  const transformOutput = glMatrix.vec3.create();
  const flippedInvertedMatrix = glMatrix.mat3.create();
  glMatrix.mat3.transpose(flippedInvertedMatrix, transformMatrix);
  glMatrix.mat3.invert(flippedInvertedMatrix, flippedInvertedMatrix);
  glMatrix.vec3.transformMat3(
    transformOutput,
    transformInput,
    flippedInvertedMatrix
  );
  return pos(transformOutput[0], transformOutput[1]);
}
function insertLineToPolygon(polygon, newLine, index) {
  const wrap = (i) => ((i + polygon.length) % polygon.length + polygon.length) % polygon.length;
  const replacePrev = [];
  let prevPoint = wrap(index - 1);
  let newPointA = null;
  while (true) {
    const lineBefore = line(
      polygon[wrap(prevPoint)],
      polygon[wrap(prevPoint + 1)]
    );
    const newPointARes = intersection(lineBefore, newLine);
    if (!newPointARes || !newPointARes.hitLine1) {
      if (prevPoint === index) {
        break;
      }
      replacePrev.unshift(wrap(prevPoint));
      prevPoint = wrap(prevPoint - 1);
    } else {
      newPointA = newPointARes.pos;
      break;
    }
  }
  const replaceNext = [];
  let nextPoint = wrap(index + 1);
  let newPointB = null;
  while (true) {
    const lineAfter = line(
      polygon[wrap(nextPoint - 1)],
      polygon[wrap(nextPoint)]
    );
    const newPointBRes = intersection(newLine, lineAfter);
    if (!newPointBRes || !newPointBRes.hitLine2) {
      if (nextPoint === index) {
        break;
      }
      replaceNext.push(wrap(nextPoint));
      nextPoint = wrap(nextPoint + 1);
    } else {
      newPointB = newPointBRes.pos;
      break;
    }
  }
  if (newPointA == null || newPointB == null) {
    newPointA = polygon[wrap(index - 1)];
    newPointB = polygon[wrap(index + 1)];
  }
  const targetPoint = wrap(index);
  const result = [...polygon];
  for (const i of replacePrev) {
    result[i] = pos(newPointA[0], newPointA[1]);
  }
  for (const i of replaceNext) {
    result[i] = pos(newPointB[0], newPointB[1]);
  }
  result.splice(
    targetPoint,
    1,
    pos(newPointA[0], newPointA[1]),
    pos(newPointB[0], newPointB[1])
  );
  return result;
}
function toClipPath(poss, unit) {
  return `polygon(${poss.map(
    (i) => `calc(${unit} * ${toCSSNumber(
      i[0]
    )}) calc(${unit} * ${toCSSNumber(i[1])})`
  ).join(", ")})`;
}
const SHADOW_SIZE_TOLERANCE = 1.3;
const offsetMatrix = (matrix, x, y) => {
  const res = glMatrix.mat3.create();
  const matrixTranslate = glMatrix.mat3.fromTranslation(
    glMatrix.mat3.create(),
    glMatrix.vec2.fromValues(x, y)
  );
  glMatrix.mat3.transpose(matrixTranslate, matrixTranslate);
  const matrixInverted = glMatrix.mat3.fromTranslation(
    glMatrix.mat3.create(),
    glMatrix.vec2.fromValues(-x, -y)
  );
  glMatrix.mat3.transpose(matrixInverted, matrixInverted);
  glMatrix.mat3.mul(res, matrixTranslate, matrix);
  glMatrix.mat3.mul(res, res, matrixInverted);
  return res;
};
const formatStyle = (matrixFlip, polygonRemain, boxShadowWidth, polygonFlip, matrixFlipShadow, polygonFlipShadow, matrixEffect, polygonEffect, width, height, unit) => {
  const transformFlip = toCSSMatrix(
    offsetMatrix(matrixFlip, width / 2, height / 2),
    unit
  );
  const clipPathRemain = toClipPath(polygonRemain, unit);
  const boxShadow = `0px 0px calc(${unit} * ${toCSSNumber(
    boxShadowWidth
  )}) 0px rgba(0, 0, 0, 1)`;
  const clipPathFlip = toClipPath(polygonFlip, unit);
  const transformFlipShadow = toCSSMatrix(
    offsetMatrix(matrixFlipShadow, width / 2, height / 2),
    unit
  );
  const clipPathFlipShadow = toClipPath(polygonFlipShadow, unit);
  const transformEffect = toCSSMatrix(
    offsetMatrix(matrixEffect, width / 2, height / 2),
    unit
  );
  const clipPathEffect = toClipPath(polygonEffect, unit);
  return {
    flipFront: {
      clipPath: clipPathRemain
    },
    flipBack: {
      transform: transformFlip,
      clipPath: clipPathFlip
    },
    flipEffect: matrixEffect.includes(NaN) ? {
      display: "none"
    } : {
      display: "",
      transform: transformEffect,
      clipPath: clipPathEffect,
      background: "linear-gradient(to right, rgba(0, 0, 0, 0.5), rgba(255, 255, 255, 0.3) 30%, rgba(0, 0, 0, 0))"
    },
    flipShadow: {
      boxShadow,
      transform: transformFlipShadow,
      clipPath: clipPathFlipShadow
    }
  };
};
const getEffectLeftTop = (width, height, leftOffset, topOffset, maxShadowWidth, unit = "1px") => {
  const posTop = pos(topOffset, 0);
  const posLeft = pos(0, leftOffset);
  const matrixFlip = genReflectMatrix(posTop, posLeft);
  const polygonRemain = [
    posTop,
    posLeft,
    pos(0, height),
    pos(width, height),
    pos(width, 0)
  ];
  const boxShadowWidth = Math.min(leftOffset, topOffset, maxShadowWidth);
  const polygonFlip = [posTop, posLeft, pos(0, 0)];
  let polygonFlipShadow = leftOffset >= topOffset ? expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(polygonFlip, [2], boxShadowWidth * SHADOW_SIZE_TOLERANCE);
  polygonFlipShadow = insertLineToPolygon(
    polygonFlipShadow,
    line(
      polygonFlip[0],
      applyReverseTransform(
        posAdd(applyReverseTransform(polygonFlip[0], matrixFlip), pos(10, 0)),
        matrixFlip
      )
    ),
    0
  );
  if (leftOffset >= topOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      2
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  }
  const posVector = glMatrix.vec2.fromValues(-1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(posTop[1] - posLeft[1], posLeft[0] - posTop[0]);
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, 0),
    applyTransform(posTop, matrixFlip),
    applyTransform(posLeft, matrixFlip),
    applyTransform(posAdd(posTop, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const getEffectRightTop = (width, height, rightOffset, topOffset, maxShadowWidth, unit = "1px") => {
  const posTop = pos(topOffset, 0);
  const posRight = pos(width, rightOffset);
  const matrixFlip = genReflectMatrix(posTop, posRight);
  const polygonRemain = [
    posTop,
    posRight,
    pos(width, height),
    pos(0, height),
    pos(0, 0)
  ];
  const boxShadowWidth = Math.min(
    rightOffset,
    width - topOffset,
    maxShadowWidth
  );
  const polygonFlip = [posTop, posRight, pos(width, 0)];
  let polygonFlipShadow = rightOffset >= width - topOffset ? expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(polygonFlip, [2], boxShadowWidth * SHADOW_SIZE_TOLERANCE);
  polygonFlipShadow = insertLineToPolygon(
    polygonFlipShadow,
    line(
      polygonFlip[0],
      applyReverseTransform(
        posAdd(applyReverseTransform(polygonFlip[0], matrixFlip), pos(10, 0)),
        matrixFlip
      )
    ),
    0
  );
  if (rightOffset >= width - topOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      2
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  }
  const posVector = glMatrix.vec2.fromValues(1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(posTop[1] - posRight[1], posRight[0] - posTop[0]);
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, 0),
    applyTransform(posTop, matrixFlip),
    applyTransform(posRight, matrixFlip),
    applyTransform(posAdd(posTop, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const getEffectLeft = (width, height, topOffset, bottomOffset, maxShadowWidth, unit = "1px") => {
  if (bottomOffset === 0) {
    return getEffectLeftTop(width, height, height, topOffset, maxShadowWidth);
  }
  if (topOffset === 0) {
    return getEffectLeftBottom(width, height, 0, bottomOffset, maxShadowWidth);
  }
  const posTop = pos(topOffset, 0);
  const posBottom = pos(bottomOffset, height);
  const matrixFlip = genReflectMatrix(posTop, posBottom);
  const polygonRemain = [posTop, pos(width, 0), pos(width, height), posBottom];
  const boxShadowWidth = Math.min(
    Math.max(topOffset, bottomOffset),
    Math.max(width - topOffset, width - bottomOffset),
    maxShadowWidth
  );
  const polygonFlip = [posTop, pos(0, 0), pos(0, height), posBottom];
  let polygonFlipShadow = topOffset > bottomOffset ? expandPolygon(
    polygonFlip,
    [0, 1],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  );
  if (topOffset > bottomOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[0],
        applyReverseTransform(
          pos(polygonFlip[0][0] + 10, polygonFlip[0][1]),
          matrixFlip
        )
      ),
      0
    );
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[3],
        applyReverseTransform(
          pos(polygonFlip[3][0] + 10, polygonFlip[3][1]),
          matrixFlip
        )
      ),
      3
    );
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      1
    );
  }
  const posVector = glMatrix.vec2.fromValues(-1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(
    posTop[1] - posBottom[1],
    posBottom[0] - posTop[0]
  );
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, 0),
    applyTransform(posTop, matrixFlip),
    applyTransform(posBottom, matrixFlip),
    applyTransform(posAdd(posTop, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const getEffectRight = (width, height, topOffset, bottomOffset, maxShadowWidth, unit = "1px") => {
  if (bottomOffset === width) {
    return getEffectRightTop(width, height, height, topOffset, maxShadowWidth);
  }
  if (topOffset === width) {
    return getEffectRightBottom(width, height, 0, bottomOffset, maxShadowWidth);
  }
  const posTop = pos(topOffset, 0);
  const posBottom = pos(bottomOffset, height);
  const matrixFlip = genReflectMatrix(posTop, posBottom);
  const polygonRemain = [posTop, pos(0, 0), pos(0, height), posBottom];
  const boxShadowWidth = Math.min(
    Math.max(width - topOffset, width - bottomOffset),
    Math.max(topOffset, bottomOffset),
    maxShadowWidth
  );
  const polygonFlip = [posTop, pos(width, 0), pos(width, height), posBottom];
  let polygonFlipShadow = width - topOffset > width - bottomOffset ? expandPolygon(
    polygonFlip,
    [0, 1],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  );
  if (width - topOffset > width - bottomOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[0],
        applyReverseTransform(
          pos(polygonFlip[0][0] + 10, polygonFlip[0][1]),
          matrixFlip
        )
      ),
      0
    );
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[3],
        applyReverseTransform(
          pos(polygonFlip[3][0] + 10, polygonFlip[3][1]),
          matrixFlip
        )
      ),
      3
    );
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      1
    );
  }
  const posVector = glMatrix.vec2.fromValues(1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(
    posTop[1] - posBottom[1],
    posBottom[0] - posTop[0]
  );
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, 0),
    applyTransform(posTop, matrixFlip),
    applyTransform(posBottom, matrixFlip),
    applyTransform(posAdd(posTop, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const getEffectLeftBottom = (width, height, leftOffset, bottomOffset, maxShadowWidth, unit = "1px") => {
  const posBottom = pos(bottomOffset, height);
  const posLeft = pos(0, leftOffset);
  const matrixFlip = genReflectMatrix(posBottom, posLeft);
  const polygonRemain = [
    posBottom,
    posLeft,
    pos(0, 0),
    pos(width, 0),
    pos(width, height)
  ];
  const boxShadowWidth = Math.min(
    height - leftOffset,
    bottomOffset,
    maxShadowWidth
  );
  const polygonFlip = [posBottom, posLeft, pos(0, height)];
  let polygonFlipShadow = height - leftOffset >= bottomOffset ? expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(polygonFlip, [2], boxShadowWidth * SHADOW_SIZE_TOLERANCE);
  polygonFlipShadow = insertLineToPolygon(
    polygonFlipShadow,
    line(
      polygonFlip[0],
      applyReverseTransform(
        posAdd(applyReverseTransform(polygonFlip[0], matrixFlip), pos(10, 0)),
        matrixFlip
      )
    ),
    0
  );
  if (height - leftOffset >= bottomOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      2
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  }
  const posVector = glMatrix.vec2.fromValues(-1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(
    posLeft[1] - posBottom[1],
    posBottom[0] - posLeft[0]
  );
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, height),
    applyTransform(posLeft, matrixFlip),
    applyTransform(posBottom, matrixFlip),
    applyTransform(posAdd(posBottom, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const getEffectRightBottom = (width, height, rightOffset, bottomOffset, maxShadowWidth, unit = "1px") => {
  const posBottom = pos(bottomOffset, height);
  const posRight = pos(width, rightOffset);
  const matrixFlip = genReflectMatrix(posBottom, posRight);
  const polygonRemain = [
    posBottom,
    posRight,
    pos(width, 0),
    pos(0, 0),
    pos(0, height)
  ];
  const boxShadowWidth = Math.min(
    height - rightOffset,
    width - bottomOffset,
    maxShadowWidth
  );
  const polygonFlip = [posBottom, posRight, pos(width, height)];
  let polygonFlipShadow = height - rightOffset >= width - bottomOffset ? expandPolygon(
    polygonFlip,
    [1, 2],
    boxShadowWidth * SHADOW_SIZE_TOLERANCE
  ) : expandPolygon(polygonFlip, [2], boxShadowWidth * SHADOW_SIZE_TOLERANCE);
  polygonFlipShadow = insertLineToPolygon(
    polygonFlipShadow,
    line(
      polygonFlip[0],
      applyReverseTransform(
        posAdd(applyReverseTransform(polygonFlip[0], matrixFlip), pos(10, 0)),
        matrixFlip
      )
    ),
    0
  );
  if (height - rightOffset >= width - bottomOffset) {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[1],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[1], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      2
    );
  } else {
    polygonFlipShadow = insertLineToPolygon(
      polygonFlipShadow,
      line(
        polygonFlip[2],
        applyReverseTransform(
          posAdd(applyReverseTransform(polygonFlip[2], matrixFlip), pos(10, 0)),
          matrixFlip
        )
      ),
      3
    );
  }
  const posVector = glMatrix.vec2.fromValues(1, 0);
  glMatrix.vec2.normalize(posVector, posVector);
  const shadowDirection = pos(
    posRight[1] - posBottom[1],
    posBottom[0] - posRight[0]
  );
  glMatrix.vec2.normalize(shadowDirection, shadowDirection);
  const lengthRatio = 1 / (Math.abs(
    shadowDirection[0] * posVector[0] + shadowDirection[1] * posVector[1]
  ) / (shadowDirection[0] * shadowDirection[0] + shadowDirection[1] * shadowDirection[1]) ** 0.5);
  glMatrix.vec2.scale(posVector, posVector, boxShadowWidth * lengthRatio);
  const matrixEffect = genMatrix(
    pos(0, 0),
    pos(0, height),
    pos(width, height),
    applyTransform(posRight, matrixFlip),
    applyTransform(posBottom, matrixFlip),
    applyTransform(posAdd(posBottom, posVector), matrixFlip)
  );
  const polygonEffect = polygonFlip.map((i) => {
    return applyTransform(i, matrixFlip);
  }).map((i) => {
    return applyReverseTransform(i, matrixEffect);
  });
  const mirror = genReflectMatrix([width / 2, 0], [width / 2, height]);
  const mirroredMatrixFlip = glMatrix.mat3.create();
  glMatrix.mat3.mul(mirroredMatrixFlip, mirror, matrixFlip);
  const mirroredPolygonFlip = polygonFlip.map((p) => {
    return applyTransform(p, mirror);
  });
  return formatStyle(
    mirroredMatrixFlip,
    polygonRemain,
    boxShadowWidth,
    mirroredPolygonFlip,
    matrixFlip,
    polygonFlipShadow,
    matrixEffect,
    polygonEffect,
    width,
    height,
    unit
  );
};
const createEffectLeft = (width, height, centerX, centerY, angle, maxShadowWidth, unit = "1px") => {
  const angleLine = Math.PI / 2 + angle;
  const vector = pos(Math.cos(angleLine), Math.sin(angleLine));
  const lineStart = pos(centerX, centerY);
  const lineTo = posAdd(lineStart, vector);
  const lineSplit = line(lineStart, lineTo);
  const leftTop = pos(0, 0);
  const leftBottom = pos(0, height);
  const rightTop = pos(width, 0);
  const rightBottom = pos(width, height);
  const lineTop = line(leftTop, rightTop);
  const lineRight = line(rightTop, rightBottom);
  const lineBottom = line(rightBottom, leftBottom);
  const lineLeft = line(leftBottom, leftTop);
  const resTop = intersection(lineTop, lineSplit);
  const resRight = intersection(lineRight, lineSplit);
  const resBottom = intersection(lineBottom, lineSplit);
  const resLeft = intersection(lineLeft, lineSplit);
  if ((resTop == null ? void 0 : resTop.hitLine1) && (resBottom == null ? void 0 : resBottom.hitLine1)) {
    return getEffectLeft(
      width,
      height,
      resTop.pos[0],
      resBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resLeft == null ? void 0 : resLeft.hitLine1) && (resRight == null ? void 0 : resRight.hitLine1)) {
    if (resLeft.pos[1] < resRight.pos[1]) {
      const p2 = glMatrix.vec2.equals(pos(width, height), lineStart) ? pos(0, 0) : pos(width, height);
      const newLine = line(lineStart, p2);
      const newResTop = intersection(lineTop, newLine);
      const newResLeft = intersection(lineLeft, newLine);
      if (newResTop == null ? void 0 : newResTop.hitLine1) {
        return getEffectLeft(
          width,
          height,
          newResTop.pos[0],
          width,
          maxShadowWidth,
          unit
        );
      } else {
        return getEffectLeftBottom(
          width,
          height,
          newResLeft.pos[1],
          width,
          maxShadowWidth,
          unit
        );
      }
    } else {
      const p2 = glMatrix.vec2.equals(pos(width, 0), lineStart) ? pos(0, height) : pos(width, 0);
      const newLine = line(lineStart, p2);
      const newResBottom = intersection(lineBottom, newLine);
      const newResLeft = intersection(lineLeft, newLine);
      if (newResBottom.hitLine1) {
        return getEffectLeft(
          width,
          height,
          width,
          newResBottom.pos[0],
          maxShadowWidth,
          unit
        );
      } else {
        return getEffectLeftTop(
          width,
          height,
          newResLeft.pos[1],
          width,
          maxShadowWidth,
          unit
        );
      }
    }
  } else if ((resLeft == null ? void 0 : resLeft.hitLine1) && (resTop == null ? void 0 : resTop.hitLine1)) {
    return getEffectLeftTop(
      width,
      height,
      resLeft.pos[1],
      resTop.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resLeft == null ? void 0 : resLeft.hitLine1) && (resBottom == null ? void 0 : resBottom.hitLine1)) {
    return getEffectLeftBottom(
      width,
      height,
      resLeft.pos[1],
      resBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resTop == null ? void 0 : resTop.hitLine1) && (resRight == null ? void 0 : resRight.hitLine1)) {
    const p2 = pos(width, height);
    const newLine = line(lineStart, p2);
    const newResTop = intersection(lineTop, newLine);
    return getEffectLeft(
      width,
      height,
      newResTop.pos[0],
      width,
      maxShadowWidth,
      unit
    );
  } else if ((resBottom == null ? void 0 : resBottom.hitLine1) && (resRight == null ? void 0 : resRight.hitLine1)) {
    const p2 = pos(width, 0);
    const newLine = line(lineStart, p2);
    const newResBottom = intersection(lineBottom, newLine);
    return getEffectLeft(
      width,
      height,
      width,
      newResBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else {
    console.warn("invalid style result");
    return {
      flipFront: {},
      flipBack: {},
      flipEffect: {},
      flipShadow: {}
    };
  }
};
const createEffectRight = (width, height, centerX, centerY, angle, maxShadowWidth, unit = "1px") => {
  const angleLine = Math.PI / 2 + angle;
  const vector = pos(Math.cos(angleLine), Math.sin(angleLine));
  const lineStart = pos(centerX, centerY);
  const lineTo = posAdd(lineStart, vector);
  const lineSplit = line(lineStart, lineTo);
  const leftTop = pos(0, 0);
  const leftBottom = pos(0, height);
  const rightTop = pos(width, 0);
  const rightBottom = pos(width, height);
  const lineTop = line(leftTop, rightTop);
  const lineRight = line(rightTop, rightBottom);
  const lineBottom = line(rightBottom, leftBottom);
  const lineLeft = line(leftBottom, leftTop);
  const resTop = intersection(lineTop, lineSplit);
  const resRight = intersection(lineRight, lineSplit);
  const resBottom = intersection(lineBottom, lineSplit);
  const resLeft = intersection(lineLeft, lineSplit);
  if ((resTop == null ? void 0 : resTop.hitLine1) && (resBottom == null ? void 0 : resBottom.hitLine1)) {
    return getEffectRight(
      width,
      height,
      resTop.pos[0],
      resBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resLeft == null ? void 0 : resLeft.hitLine1) && (resRight == null ? void 0 : resRight.hitLine1)) {
    if (resLeft.pos[1] > resRight.pos[1]) {
      const p2 = glMatrix.vec2.equals(pos(0, height), lineStart) ? pos(width, 0) : pos(0, height);
      const newLine = line(lineStart, p2);
      const newResTop = intersection(lineTop, newLine);
      const newResRight = intersection(lineRight, newLine);
      if (newResTop.hitLine1) {
        return getEffectRight(
          width,
          height,
          newResTop.pos[0],
          0,
          maxShadowWidth,
          unit
        );
      } else {
        return getEffectRightBottom(
          width,
          height,
          newResRight.pos[1],
          0,
          maxShadowWidth,
          unit
        );
      }
    } else {
      const p2 = glMatrix.vec2.equals(pos(0, 0), lineStart) ? pos(width, height) : pos(0, 0);
      const newLine = line(lineStart, p2);
      const newResBottom = intersection(lineBottom, newLine);
      const newResRight = intersection(lineRight, newLine);
      if (newResBottom.hitLine1) {
        return getEffectRight(
          width,
          height,
          0,
          newResBottom.pos[0],
          maxShadowWidth,
          unit
        );
      } else {
        return getEffectRightTop(
          width,
          height,
          newResRight.pos[1],
          0,
          maxShadowWidth,
          unit
        );
      }
    }
  } else if ((resRight == null ? void 0 : resRight.hitLine1) && (resTop == null ? void 0 : resTop.hitLine1)) {
    return getEffectRightTop(
      width,
      height,
      resRight.pos[1],
      resTop.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resRight == null ? void 0 : resRight.hitLine1) && (resBottom == null ? void 0 : resBottom.hitLine1)) {
    return getEffectRightBottom(
      width,
      height,
      resRight.pos[1],
      resBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else if ((resTop == null ? void 0 : resTop.hitLine1) && (resLeft == null ? void 0 : resLeft.hitLine1)) {
    const p2 = pos(0, height);
    const newLine = line(lineStart, p2);
    const newResTop = intersection(lineTop, newLine);
    return getEffectRight(
      width,
      height,
      newResTop.pos[0],
      0,
      maxShadowWidth,
      unit
    );
  } else if ((resBottom == null ? void 0 : resBottom.hitLine1) && (resLeft == null ? void 0 : resLeft.hitLine1)) {
    const p2 = pos(0, 0);
    const newLine = line(lineStart, p2);
    const newResBottom = intersection(lineBottom, newLine);
    return getEffectRight(
      width,
      height,
      0,
      newResBottom.pos[0],
      maxShadowWidth,
      unit
    );
  } else {
    console.warn("invalid style result");
    return {
      flipFront: {
        clipPath: "none"
      },
      flipBack: {
        transform: "none",
        clipPath: "none"
      },
      flipEffect: {
        transform: "none",
        clipPath: "none"
      },
      flipShadow: {
        boxShadow: "none",
        transform: "none",
        clipPath: "none"
      }
    };
  }
};
export {
  createEffectLeft,
  createEffectRight,
  getEffectLeft,
  getEffectLeftBottom,
  getEffectLeftTop,
  getEffectRight,
  getEffectRightBottom,
  getEffectRightTop
};
//# sourceMappingURL=simple-page-flip.js.map
